@@ -34,6 +34,8 @@ class ChatEdit : public KChatEdit
         void triggerCompletion();
         void cancelCompletion();
 
+        void insertMention(QString author);
+
     signals:
         void proposedCompletion(const QStringList& allCompletions, int curIndex);
         void cancelledCompletion();
@@ -48,6 +50,8 @@ class ChatEdit : public KChatEdit
         QStringList completionMatches;
         int matchesListPosition;
 
+        bool pickingMentions = false;
+
         void startNewCompletion();
         void appendTextAtCursor(const QString& text, bool select = false);
 };
