@@ -1,3 +1,5 @@
+#include <utility>
+
 /**************************************************************************
  *                                                                        *
  * Copyright (C) 2015 Felix Rohrbach <kde@fxrh.de>                        *
@@ -21,11 +23,16 @@
 
 #include <QtWidgets/QVBoxLayout>
 #include <QtWidgets/QLabel>
+#include <QtWidgets/QFileDialog>
+#include <QtWidgets/QApplication>
 
 #include <QtQml/QQmlContext>
 #include <QtQml/QQmlEngine>
+#ifdef DISABLE_QQUICKWIDGET
 #include <QtQuick/QQuickView>
-#include <QtQuick/QQuickItem>
+#else
+#include <QtQuickWidgets/QQuickWidget>
+#endif
 #include <QtCore/QRegularExpression>
 #include <QtCore/QStringBuilder>
 
@@ -37,14 +44,23 @@
 #include "imageprovider.h"
 #include "chatedit.h"
 
+static const auto DefaultPlaceholderText =
+        ChatRoomWidget::tr("Choose a room to send messages or enter a command...");
+
 ChatRoomWidget::ChatRoomWidget(QWidget* parent)
     : QWidget(parent)
+    , m_messageModel(new MessageEventModel(this))
     , m_currentRoom(nullptr)
     , readMarkerOnScreen(false)
 {
-    qmlRegisterType<QuaternionRoom>();
-    qmlRegisterType<QMatrixClient::Settings>("QMatrixClient", 1, 0, "Settings");
-    m_messageModel = new MessageEventModel(this);
+    {
+        using namespace QMatrixClient;
+        qmlRegisterUncreatableType<QuaternionRoom>("QMatrixClient", 1, 0, "Room",
+            "Room objects can only be created by libqmatrixclient");
+        qmlRegisterUncreatableType<User>("QMatrixClient", 1, 0, "User",
+            "User objects can only be created by libqmatrixclient");
+        qmlRegisterType<Settings>("QMatrixClient", 1, 0, "Settings");
+    }
 
     m_roomAvatar = new QLabel();
     m_roomAvatar->setPixmap({});
@@ -60,25 +76,34 @@ ChatRoomWidget::ChatRoomWidget(QWidget* parent)
     auto topicSeparator = new QFrame();
     topicSeparator->setFrameShape(QFrame::HLine);
 
-    m_quickView = new QQuickView();
+    m_timelineWidget = new timelineWidget_t;
+#ifdef DISABLE_QQUICKWIDGET
+    qDebug() << "Using QQuickView wrapper instead of QQuickWidget";
+    auto* container = QWidget::createWindowContainer(m_timelineWidget, this);
+    container->
+#else
+    qDebug() << "Using QQuickWidget to render QML";
+    m_timelineWidget->
+#endif // Use different objects but the same method with the same parameters
+            setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
+
+    m_timelineWidget->setResizeMode(timelineWidget_t::SizeRootObjectToView);
 
     m_imageProvider = new ImageProvider(nullptr); // No connection yet
-    m_quickView->engine()->addImageProvider("mtx", m_imageProvider);
+    m_timelineWidget->engine()->addImageProvider("mtx", m_imageProvider);
 
-    QWidget* container = QWidget::createWindowContainer(m_quickView, this);
-    container->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
-    QQmlContext* ctxt = m_quickView->rootContext();
+    QQmlContext* ctxt = m_timelineWidget->rootContext();
     ctxt->setContextProperty("messageModel", m_messageModel);
     ctxt->setContextProperty("controller", this);
     ctxt->setContextProperty("debug", QVariant(false));
-    m_quickView->setSource(QUrl("qrc:///qml/chat.qml"));
-    m_quickView->setResizeMode(QQuickView::SizeRootObjectToView);
+
+    m_timelineWidget->setSource(QUrl("qrc:///qml/Timeline.qml"));
 
     m_currentlyTyping = new QLabel();
     m_currentlyTyping->setWordWrap(true);
 
     m_chatEdit = new ChatEdit(this);
-    m_chatEdit->setPlaceholderText(tr("Send a message (unencrypted)..."));
+    m_chatEdit->setPlaceholderText(DefaultPlaceholderText);
     m_chatEdit->setAcceptRichText(false);
     connect( m_chatEdit, &KChatEdit::returnPressed, this, &ChatRoomWidget::sendInput );
     connect(m_chatEdit, &ChatEdit::proposedCompletion, this,
@@ -99,7 +124,13 @@ ChatRoomWidget::ChatRoomWidget(QWidget* parent)
     layout->addLayout(headerLayout);
 
     layout->addWidget(topicSeparator);
-    layout->addWidget(container);
+    layout->addWidget(
+#ifdef DISABLE_QQUICKWIDGET
+                container
+#else
+                m_timelineWidget
+#endif
+    );
     layout->addWidget(m_currentlyTyping);
     layout->addWidget(m_chatEdit);
     setLayout(layout);
@@ -107,7 +138,7 @@ ChatRoomWidget::ChatRoomWidget(QWidget* parent)
 
 void ChatRoomWidget::enableDebug()
 {
-    QQmlContext* ctxt = m_quickView->rootContext();
+    QQmlContext* ctxt = m_timelineWidget->rootContext();
     ctxt->setContextProperty("debug", true);
 }
 
@@ -121,7 +152,6 @@ void ChatRoomWidget::setRoom(QuaternionRoom* room)
     if( m_currentRoom )
     {
         m_currentRoom->setCachedInput( m_chatEdit->toPlainText() );
-        m_currentRoom->setShown(false);
         roomHistories.insert(m_currentRoom, m_chatEdit->history());
         m_currentRoom->connection()->disconnect(this);
         m_currentRoom->disconnect( this );
@@ -132,6 +162,7 @@ void ChatRoomWidget::setRoom(QuaternionRoom* room)
     m_chatEdit->cancelCompletion();
 
     m_currentRoom = room;
+    m_timelineWidget->rootContext()->setContextProperty("room", room);
     if( m_currentRoom )
     {
         using namespace QMatrixClient;
@@ -157,33 +188,21 @@ void ChatRoomWidget::setRoom(QuaternionRoom* room)
             reStartShownTimer();
             emit readMarkerMoved();
         });
+        connect( m_currentRoom, &Room::encryption,
+                 this, &ChatRoomWidget::encryptionChanged);
         connect(m_currentRoom->connection(), &Connection::loggedOut,
                 this, [this]
         {
             qWarning() << "Logged out, escaping the room";
             setRoom(nullptr);
         });
-        connect(m_currentRoom->connection(), &Connection::joinedRoom,
-                this, [this] (Room* r, Room* prev)
-        {
-            if (m_currentRoom == prev)
-                setRoom(static_cast<QuaternionRoom*>(r));
-        });
-        connect(m_currentRoom->connection(), &Connection::aboutToDeleteRoom,
-                this, [this] (Room* r)
-        {
-            if (m_currentRoom == r)
-                setRoom(nullptr);
-        });
-
-        m_currentRoom->setShown(true);
     } else
         m_imageProvider->setConnection(nullptr);
     updateHeader();
     typingChanged();
+    encryptionChanged();
+
     m_messageModel->changeRoom( m_currentRoom );
-    QObject* rootItem = m_quickView->rootObject();
-    QMetaObject::invokeMethod(rootItem, "scrollToBottom");
 }
 
 void ChatRoomWidget::typingChanged()
@@ -212,7 +231,8 @@ void ChatRoomWidget::updateHeader()
         m_topicLabel->setText("<strong>" % m_currentRoom->name() %
                               "</strong><br />" % prettyTopic);
         auto avatarSize = m_topicLabel->heightForWidth(width());
-        m_roomAvatar->setPixmap(m_currentRoom->avatar(avatarSize, avatarSize));
+        m_roomAvatar->setPixmap(
+                QPixmap::fromImage(m_currentRoom->avatar(avatarSize)));
     }
     else
     {
@@ -221,162 +241,250 @@ void ChatRoomWidget::updateHeader()
     }
 }
 
-bool ChatRoomWidget::checkAndRun(const QString& checkArg, const QString& pattern,
-                                 std::function<void()> fn,
-                                 const QString& errorMsg) const
+void ChatRoomWidget::encryptionChanged()
 {
-    if (QRegularExpression(pattern).match(checkArg).hasMatch())
+    m_chatEdit->setReadOnly(m_currentRoom && m_currentRoom->usesEncryption());
+    m_chatEdit->setPlaceholderText(
+        m_currentRoom
+            ? m_currentRoom->usesEncryption()
+                ? tr("Sending encrypted messages is not supported yet")
+                : tr("Send a message (unencrypted) or enter a command...")
+            : DefaultPlaceholderText);
+}
+
+void ChatRoomWidget::insertMention(QString author)
+{
+    m_chatEdit->insertMention(std::move(author));
+}
+
+void ChatRoomWidget::focusInput()
+{
+    m_chatEdit->setFocus();
+}
+
+/**
+ * \brief Split the string into no more than specified numbers of parts
+ * The function takes \p s and splits it into at most \p maxParts parts
+ * using \p sep for the separator. Empty parts are skipped. If there are more
+ * than \p maxParts parts in the string, the last returned part includes
+ * the remainder of the string.
+ * \return the vector of references to the original string, one reference for
+ * each part.
+ */
+QVector<QString> lazySplitRef(const QString& s, QChar sep, int maxParts)
+{
+    QVector<QString> parts;
+    int pos = 0, nextPos = 0;
+    for (; maxParts > 1 && (nextPos = s.indexOf(sep, pos)) > -1; --maxParts)
     {
-        fn();
-        return true;
+        parts.push_back(s.mid(pos, nextPos - pos));
+        while (s[++nextPos] == sep)
+            ;
+        pos = nextPos;
     }
-
-    emit showStatusMessage(errorMsg, 5000);
-    return false;
+    parts.push_back(s.mid(pos));
+    return parts;
 }
 
-bool ChatRoomWidget::checkAndRun1(const QString& args, const QString& pattern,
-    std::function<void(QMatrixClient::Room*, QString)> fn1,
-    const QString& errorMsg) const
-{
-    return checkAndRun(args, pattern,
-                       std::bind(fn1, m_currentRoom, args), errorMsg);
-}
-
-bool ChatRoomWidget::checkAndRun2(const QString& args, const QString& pattern1,
-    std::function<void(QMatrixClient::Room*, QString, QString)> fn2,
-    const QString& errorMsg) const
-{
-    const auto nextSpacePos = args.indexOf(' ', 1);
-    const auto arg1 = args.left(nextSpacePos);
-    const auto arg2 = args.mid(nextSpacePos + 1).trimmed();
-    return checkAndRun(arg1, pattern1,
-                       std::bind(fn2, m_currentRoom, arg1, arg2), errorMsg);
-}
-
-bool ChatRoomWidget::doSendInput()
+QString ChatRoomWidget::doSendInput()
 {
     QString text = m_chatEdit->toPlainText();
     if ( text.isEmpty() )
-        return false;
+        return tr("There's nothing to send");
 
     if (!text.startsWith('/'))
     {
         m_currentRoom->postMessage(text);
-        return true;
+        return {};
     }
     if (text[1] == '/')
     {
         text.remove(0, 1);
         m_currentRoom->postMessage(text);
-        return true;
+        return {};
     }
 
+    static const auto ReFlags =
+            QRegularExpression::DotMatchesEverythingOption|
+            QRegularExpression::OptimizeOnFirstUsageOption;
+
+    static const QRegularExpression
+            CommandRe { "^/([^ ]+)( +(.*))?\\s*$", ReFlags },
+            RoomIdRE { "^[#!][-0-9a-z._=]+:\\S+$", ReFlags },
+            UserIdRE { "^@[-0-9a-z._=]+:\\S+$", ReFlags };
+
     // Process a command
-    static const QRegularExpression re("^/([^ ]+)( (.*))?$");
-    const auto matches = re.match(text);
+    const auto matches = CommandRe.match(text);
     const auto command = matches.capturedRef(1);
-    const auto args = matches.captured(3);
+    const auto argString = matches.captured(3);
 
-    static const auto ROOM_ID = QStringLiteral("^[#!][-0-9a-z._=]+:.+$");
-    static const auto USER_ID = QStringLiteral("^@[-0-9a-z._=]+:.+$");
-
-    using QMatrixClient::Room;
     // Commands available without a current room
     if (command == "join")
     {
-        return checkAndRun(args, ROOM_ID,
-            [=] { emit joinCommandEntered(args); },
-            tr("/join argument doesn't look like a room ID or alias"));
+        if (!argString.contains(RoomIdRE))
+            return tr("/join argument doesn't look like a room ID or alias");
+        emit joinCommandEntered(argString);
+        return {};
+    }
+    if (command == "quit")
+    {
+        qApp->closeAllWindows();
+        return {};
     }
     // --- Add more roomless commands here
     if (!m_currentRoom)
-    {
-        emit showStatusMessage(
-            tr("There's no such /command outside of room."
-                   " Start with // to send this line literally"), 5000);
-        return false;
-    }
+        return tr("There's no such /command outside of room."
+                  " Start with // to send this line literally");
+
     // Commands available only in the room context
-    if (command == "leave")
+    using QMatrixClient::Room;
+    if (command == "leave" || command == "part")
     {
+        if (!argString.isEmpty())
+            return tr("Sending a farewell message is not supported yet."
+                      " If you intended to leave another room, switch to it"
+                      " and type /leave there.");
+
         m_currentRoom->leaveRoom();
-        return true;
+        return {};
     }
     if (command == "forget")
     {
-        if (args.isEmpty())
-        {
-            emit showStatusMessage(
-                tr("/forget should be followed by the room id,"
-                   " even for the current room"));
-            return false;
-        }
-        return checkAndRun(args, ROOM_ID,
-            [=] { m_currentRoom->connection()->forgetRoom(args); },
-            tr("/forget should be followed by the room id"));
+        if (argString.isEmpty())
+            return tr("/forget must be followed by the room id/alias,"
+                      " even for the current room");
+        if (!argString.contains(RoomIdRE))
+            return tr("%1 doesn't look like a room id or alias").arg(argString);
+
+        // Forget the specified room using the current room's connection
+        m_currentRoom->connection()->forgetRoom(argString);
+        return {};
     }
     if (command == "invite")
     {
-        return checkAndRun1(args, USER_ID, &Room::inviteToRoom,
-            tr("/invite argument doesn't look like a user ID"));
+        if (argString.isEmpty())
+            return tr("/invite <memberId>");
+        if (!argString.contains(UserIdRE))
+            return tr("%1 doesn't look like a user ID").arg(argString);
+
+        m_currentRoom->inviteToRoom(argString);
+        return {};
     }
-    if (command == "kick")
+    if (command == "kick" || command == "ban")
     {
-        return checkAndRun2(args, USER_ID, &Room::kickMember,
-            tr("The first /kick argument doesn't look like a user ID"));
-    }
-    if (command == "ban")
-    {
-        return checkAndRun2(args, USER_ID, &Room::ban,
-            tr("The first /ban argument doesn't look like a user ID"));
+        const auto args = lazySplitRef(argString, ' ', 2);
+        if (args.front().isEmpty())
+            return tr("/%1 <memberId> <reason>").arg(command.toString());
+        if (!UserIdRE.match(args.front()).hasMatch())
+            return tr("%1 doesn't look like a user id")
+                    .arg(args.front());
+
+        auto* user = m_currentRoom->user(args.front());
+        if (m_currentRoom->memberJoinState(user)
+                != QMatrixClient::JoinState::Join)
+            return tr("%1 is not a member of this room")
+                    .arg(user->fullName(m_currentRoom));
+
+        if (command == "ban")
+            m_currentRoom->ban(user->id(), args.back());
+        else
+            m_currentRoom->kickMember(user->id(), args.back());
+        return {};
     }
     if (command == "unban")
     {
-        return checkAndRun1(args, USER_ID, &Room::unban,
-            tr("/unban argument doesn't look like a user ID"));
+        if (argString.isEmpty())
+            return tr("/unban <memberId>");
+        if (!argString.contains(UserIdRE))
+            return tr("/unban argument doesn't look like a user ID");
+
+        m_currentRoom->unban(argString);
+        return {};
     }
     using MsgType = QMatrixClient::RoomMessageEvent::MsgType;
     if (command == "me")
     {
-        m_currentRoom->postMessage(args, MsgType::Emote);
-        return true;
+        if (argString.isEmpty())
+            return tr("/me needs an argument");
+        m_currentRoom->postMessage(argString, MsgType::Emote);
+        return {};
     }
     if (command == "notice")
     {
-        m_currentRoom->postMessage(args, MsgType::Notice);
-        return true;
+        if (argString.isEmpty())
+            return tr("/notice needs an argument");
+        m_currentRoom->postMessage(argString, MsgType::Notice);
+        return {};
     }
     if (command == "shrug") // Peeked at Discord
     {
         m_currentRoom->postMessage("¯\\_(ツ)_/¯");
-        return true;
+        return {};
     }
     if (command == "topic")
     {
-        m_currentRoom->setTopic(args);
-        return true;
+        m_currentRoom->setTopic(argString);
+        return {};
     }
     if (command == "nick")
     {
-        // Technically, it's legitimate to change the displayname outside of
-        // any room; however, since Quaternion allows having several connections,
-        // it needs to understand which connection to engage, and there's no good
-        // way so far to determine the connection outside of a room.
-        m_currentRoom->localUser()->rename(args);
-        return true;
+        m_currentRoom->localUser()->rename(argString);
+        return {};
+    }
+    if (command == "roomnick")
+    {
+        m_currentRoom->localUser()->rename(argString, m_currentRoom);
+        return {};
+    }
+    if (command == "pm" || command == "msg")
+    {
+        const auto args = lazySplitRef(argString, ' ', 2);
+        if (args.front().isEmpty())
+            return tr("/%1 <memberId> <message>").arg(command.toString());
+        if (RoomIdRE.match(args.front()).hasMatch() && command == "msg")
+        {
+            if (auto* room = m_currentRoom->connection()->room(args.front()))
+            {
+                room->postMessage(args.back());
+                return {};
+            }
+            else
+                return tr("%1 doesn't seem to have joined room %2")
+                        .arg(m_currentRoom->localUser()->id(),
+                             args.front());
+        }
+        if (UserIdRE.match(args.front()).hasMatch())
+        {
+            m_currentRoom->connection()->doInDirectChat(args.front(),
+                [msg=args.back()] (Room* dc) { dc->postMessage(msg); });
+            return {};
+        }
+
+        return tr("%1 doesn't look like a user id or room alias")
+                .arg(args.front());
+    }
+    if (command == "query" || command == "dc")
+    {
+        if (argString.isEmpty())
+            return tr("/%1 <memberId>").arg(command.toString());
+        if (!argString.contains(UserIdRE))
+            return tr("%1 doesn't look like a user id").arg(argString);
+
+        m_currentRoom->connection()->requestDirectChat(argString);
+        return {};
     }
     // --- Add more room commands here
     qDebug() << "Unknown command:" << command;
-    emit showStatusMessage(
-        tr("Unknown /command. Use // to send this line literally"), 5000);
-    return false;
+    return tr("Unknown /command. Use // to send this line literally");
 }
 
 void ChatRoomWidget::sendInput()
 {
-    if (doSendInput())
+    auto result = doSendInput();
+    if (result.isEmpty())
         m_chatEdit->saveInput();
+    else
+        emit showStatusMessage(result, 5000);
 }
 
 QStringList ChatRoomWidget::findCompletionMatches(const QString& pattern) const
@@ -402,6 +510,21 @@ QStringList ChatRoomWidget::findCompletionMatches(const QString& pattern) const
     return matches;
 }
 
+void ChatRoomWidget::saveFileAs(QString eventId)
+{
+    if (!m_currentRoom)
+    {
+        qWarning()
+            << "ChatRoomWidget::saveFileAs without an active room ignored";
+        return;
+    }
+    auto fileName = QFileDialog::getSaveFileName(
+                this, tr("Save file as"),
+                m_currentRoom->fileNameToDownload(eventId));
+    if (!fileName.isEmpty())
+        m_currentRoom->downloadFile(eventId, QUrl::fromLocalFile(fileName));
+}
+
 void ChatRoomWidget::onMessageShownChanged(QString eventId, bool shown)
 {
     if (!m_currentRoom)
