@@ -19,9 +19,12 @@
 
 #include <QtWidgets/QApplication>
 #include <QtWidgets/QSplashScreen>
+#include <QtCore/QTranslator>
+#include <QtCore/QLibraryInfo>
 #include <QtCore/QCommandLineParser>
 #include <QtCore/QDebug>
 
+#include "networksettings.h"
 #include "mainwindow.h"
 #include "activitydetector.h"
 #include "lib/settings.h"
@@ -32,7 +35,7 @@ int main( int argc, char* argv[] )
     QApplication::setOrganizationName("QMatrixClient");
     QApplication::setApplicationName("quaternion");
     QApplication::setApplicationDisplayName("Quaternion");
-    QApplication::setApplicationVersion("0.0.5");
+    QApplication::setApplicationVersion("0.0.9.2");
 
     QMatrixClient::Settings::setLegacyNames("Quaternion", "quaternion");
 
@@ -44,6 +47,15 @@ int main( int argc, char* argv[] )
         QApplication::postEvent(qApp, new QEvent(QEvent::Quit));
     });
 
+    QTranslator qtTranslator;
+    qtTranslator.load(QLocale(), "qt", "_",
+                      QLibraryInfo::location(QLibraryInfo::TranslationsPath));
+    app.installTranslator(&qtTranslator);
+
+    QTranslator appTranslator;
+    appTranslator.load(QLocale(), "quaternion", ".", "i18n");
+    app.installTranslator(&appTranslator);
+
     QCommandLineParser parser;
     parser.setApplicationDescription(QApplication::translate("main", "An IM client for the Matrix protocol"));
     parser.addHelpOption();
@@ -56,17 +68,26 @@ int main( int argc, char* argv[] )
     bool debugEnabled = parser.isSet(debug);
     qDebug() << "Debug: " << debugEnabled;
 
-    QSplashScreen splash(QPixmap(":/icon.png"));
-    if (!QSettings().value("UI/suppress_splash").toBool())
-    {
-        splash.show();
-        app.processEvents();
-    }
+#if (QT_VERSION >= QT_VERSION_CHECK(5, 10, 0))
+    app.setAttribute(Qt::AA_DisableWindowContextHelpButton);
+#endif
+
+//    bool splashEnabled = !QSettings().value("UI/suppress_splash").toBool();
+//    QSplashScreen splash(QPixmap(":/icon.png"));
+//    if (splashEnabled)
+//    {
+//        splash.show();
+//        app.processEvents();
+//    }
+
+    QMatrixClient::NetworkSettings().setupApplicationProxy();
 
     MainWindow window;
     if( debugEnabled )
         window.enableDebug();
-    splash.finish(&window); // calls app.processEvents()
+
+//    if (splashEnabled)
+//        splash.finish(&window); // calls app.processEvents()
 
     ActivityDetector ad(app, window); Q_UNUSED(ad);
     qDebug() << "--- Show time!";
