@@ -19,12 +19,29 @@
 
 #include "mainwindow.h"
 
+#include "roomlistdock.h"
+#include "userlistdock.h"
+#include "chatroomwidget.h"
+#include "logindialog.h"
+#include "networkconfigdialog.h"
+#include "roomdialogs.h"
+#include "systemtrayicon.h"
+
+#include "lib/csapi/joining.h"
+#include "lib/connection.h"
+#include "lib/networkaccessmanager.h"
+#include "lib/settings.h"
+#include "lib/logging.h"
+
 #include <QtCore/QTimer>
 #include <QtCore/QDebug>
 #include <QtCore/QStandardPaths>
 #include <QtCore/QStringBuilder>
 #include <QtCore/QFileInfo>
 #include <QtCore/QDir>
+#include <QtCore/QElapsedTimer>
+#include <QtNetwork/QNetworkReply>
+#include <QtNetwork/QAuthenticator>
 #include <QtWidgets/QApplication>
 #include <QtWidgets/QMenuBar>
 #include <QtWidgets/QMessageBox>
@@ -32,21 +49,27 @@
 #include <QtWidgets/QStatusBar>
 #include <QtWidgets/QLabel>
 #include <QtWidgets/QPushButton>
+#include <QtWidgets/QDialogButtonBox>
+#include <QtWidgets/QFormLayout>
 #include <QtGui/QMovie>
 #include <QtGui/QCloseEvent>
+#include <QtGui/QDesktopServices>
 
-#include "jobs/joinroomjob.h"
-#include "lib/connection.h"
-#include "roomlistdock.h"
-#include "userlistdock.h"
-#include "chatroomwidget.h"
-#include "logindialog.h"
-#include "systemtray.h"
-#include "settings.h"
+using QMatrixClient::NetworkAccessManager;
+using QMatrixClient::AccountSettings;
 
 MainWindow::MainWindow()
 {
     Connection::setRoomType<QuaternionRoom>();
+
+    // Bind callbacks to signals from NetworkAccessManager
+
+    auto nam = NetworkAccessManager::instance();
+    connect(nam, &QNetworkAccessManager::proxyAuthenticationRequired,
+        this, &MainWindow::proxyAuthenticationRequired);
+    connect(nam, &QNetworkAccessManager::sslErrors,
+            this, &MainWindow::sslErrors);
+
     setWindowIcon(QIcon(":/icon.png"));
 
     roomListDock = new RoomListDock(this);
@@ -62,17 +85,19 @@ MainWindow::MainWindow()
     connect( chatRoomWidget, &ChatRoomWidget::showStatusMessage, statusBar(), &QStatusBar::showMessage );
 
     createMenu();
-    systemTray = new SystemTray(this);
-    systemTray->show();
+    systemTrayIcon = new SystemTrayIcon(this);
+    systemTrayIcon->show();
 
     busyIndicator = new QMovie(":/busy.gif");
     busyLabel = new QLabel(this);
     busyLabel->setMovie(busyIndicator);
-    busyLabel->hide();
     statusBar()->setSizeGripEnabled(false);
     statusBar()->addPermanentWidget(busyLabel);
-    loadSettings();
+    statusBar()->showMessage(tr("Loading..."));
+    loadSettings(); // Only GUI, account settings will be loaded in invokeLogin
 
+    busyLabel->show();
+    busyIndicator->start();
     QTimer::singleShot(0, this, SLOT(invokeLogin()));
 }
 
@@ -81,24 +106,32 @@ ChatRoomWidget* MainWindow::getChatRoomWidget() const
    return chatRoomWidget;
 }
 
+template <typename DialogT, typename... DialogArgTs>
+void summon(QPointer<DialogT>& dlg, DialogArgTs&&... dialogArgs)
+{
+    if (!dlg)
+    {
+        dlg = new DialogT(std::forward<DialogArgTs>(dialogArgs)...);
+        dlg->setModal(false);
+        dlg->setAttribute(Qt::WA_DeleteOnClose);
+    }
+
+    dlg->reactivate();
+}
+
 void MainWindow::createMenu()
 {
     // Connection menu
     connectionMenu = menuBar()->addMenu(tr("&Accounts"));
 
-#if (QT_VERSION >= QT_VERSION_CHECK(5, 6, 0))
     connectionMenu->addAction(tr("&Login..."), this, [=]{ showLoginWindow(); } );
-#else
-    connectionMenu->addAction(tr("&Login..."), this, SLOT(showLoginWindow()));
-#endif
 
     connectionMenu->addSeparator();
-    // Logout actions will be added between these two separators - see addConnection()
+    // Account submenus will be added in this place - see addConnection()
     accountListGrowthPoint = connectionMenu->addSeparator();
 
-    auto quitAction = connectionMenu->addAction(tr("&Quit"));
-    quitAction->setShortcut(QKeySequence::Quit);
-    connect( quitAction, &QAction::triggered, qApp, &QApplication::closeAllWindows );
+    connectionMenu->addAction(tr("&Quit"),
+        qApp, &QApplication::closeAllWindows, QKeySequence::Quit);
 
     // View menu
     auto viewMenu = menuBar()->addMenu(tr("&View"));
@@ -109,8 +142,121 @@ void MainWindow::createMenu()
     // Room menu
     auto roomMenu = menuBar()->addMenu(tr("&Room"));
 
-    auto joinRoomAction = roomMenu->addAction(tr("&Join Room..."));
-    connect( joinRoomAction, &QAction::triggered, this, [=]{ joinRoom(); } );
+    roomSettingsAction =
+        roomMenu->addAction(tr("Change room &settings..."), [this]
+        {
+            static QHash<QuaternionRoom*, QPointer<RoomSettingsDialog>> dlgs;
+            summon(dlgs[currentRoom], currentRoom, this);
+        });
+    roomSettingsAction->setDisabled(true);
+    roomMenu->addSeparator();
+    createRoomAction =
+        roomMenu->addAction(tr("Create &new room..."), [this]
+        {
+            static QPointer<CreateRoomDialog> dlg;
+            summon(dlg, connections, this);
+        });
+    createRoomAction->setDisabled(true);
+    roomMenu->addAction(tr("&Join room..."), [=]{ joinRoom(); } );
+    roomMenu->addSeparator();
+    roomMenu->addAction(tr("&Close current room"),
+        [this] { selectRoom(nullptr); }, QKeySequence::Close);
+
+    // Settings menu
+    auto settingsMenu = menuBar()->addMenu(tr("&Settings"));
+    using QMatrixClient::Settings;
+
+    {
+        auto notifGroup = new QActionGroup(this);
+        connect(notifGroup, &QActionGroup::triggered, this,
+            [] (QAction* notifAction)
+            {
+                notifAction->setChecked(true);
+                Settings().setValue("UI/notifications",
+                                    notifAction->data().toString());
+            });
+
+        auto noNotif = notifGroup->addAction(tr("&Highlight only"));
+        noNotif->setData(QStringLiteral("none"));
+        noNotif->setStatusTip(tr("Notifications are entirely suppressed"));
+        auto gentleNotif = notifGroup->addAction(tr("&Non-intrusive"));
+        gentleNotif->setData(QStringLiteral("non-intrusive"));
+        gentleNotif->setStatusTip(
+            tr("Show notifications but do not activate the window"));
+        auto fullNotif = notifGroup->addAction(tr("&Full"));
+        fullNotif->setData(QStringLiteral("intrusive"));
+        fullNotif->setStatusTip(
+            tr("Show notifications and activate the window"));
+
+        auto notifMenu = settingsMenu->addMenu(tr("Notifications"));
+        for (auto a: {noNotif, gentleNotif, fullNotif})
+        {
+            a->setCheckable(true);
+            notifMenu->addAction(a);
+        }
+
+        const auto curSetting = Settings().value("UI/notifications",
+                                                 fullNotif->data().toString());
+        if (curSetting == noNotif->data().toString())
+            noNotif->setChecked(true);
+        else if (curSetting == gentleNotif->data().toString())
+            gentleNotif->setChecked(true);
+        else
+            fullNotif->setChecked(true);
+    }
+    {
+        auto layoutGroup = new QActionGroup(this);
+        connect(layoutGroup, &QActionGroup::triggered, this,
+            [this] (QAction* action)
+        {
+            action->setChecked(true);
+            Settings().setValue("UI/timeline_style", action->data().toString());
+            chatRoomWidget->setRoom(nullptr);
+            chatRoomWidget->setRoom(currentRoom);
+        });
+
+        auto defaultLayout = layoutGroup->addAction(tr("Default"));
+        defaultLayout->setStatusTip(
+            tr("The layout with author labels above blocks of messages"));
+        auto xchatLayout = layoutGroup->addAction(tr("XChat"));
+        xchatLayout->setData(QStringLiteral("xchat"));
+        xchatLayout->setStatusTip(
+            tr("The layout with author labels to the left from each message"));
+
+        auto layoutMenu = settingsMenu->addMenu(tr("Timeline layout"));
+        for (auto a: {defaultLayout, xchatLayout})
+        {
+            a->setCheckable(true);
+            layoutMenu->addAction(a);
+        }
+
+        const auto curSetting = Settings().value("UI/timeline_style",
+                                                 defaultLayout->data().toString());
+        if (curSetting == xchatLayout->data().toString())
+            xchatLayout->setChecked(true);
+        else
+            defaultLayout->setChecked(true);
+    }
+    auto autoloadImages =
+        settingsMenu->addAction(tr("Load full-size images at once"),
+            [this] (bool checked)
+            {
+                Settings().setValue("UI/autoload_images", checked);
+                chatRoomWidget->setRoom(nullptr);
+                chatRoomWidget->setRoom(currentRoom);
+            });
+    autoloadImages->setStatusTip(
+        tr("Automatically download a full-size image instead of a thumbnail"));
+    autoloadImages->setCheckable(true);
+    autoloadImages->setChecked(
+        Settings().value("UI/autoload_images", true).toBool());
+
+    settingsMenu->addSeparator();
+    settingsMenu->addAction(tr("Configure &network proxy..."), [this]
+    {
+        static QPointer<NetworkConfigDialog> dlg;
+        summon(dlg, this);
+    });
 }
 
 void MainWindow::loadSettings()
@@ -133,20 +279,15 @@ void MainWindow::saveSettings() const
     sg.sync();
 }
 
-inline QString accessTokenFileName(const QMatrixClient::AccountSettings& account)
+inline QString accessTokenFileName(const AccountSettings& account)
 {
     QString fileName = account.userId();
     fileName.replace(':', '_');
-    return QStandardPaths::writableLocation(
-#if (QT_VERSION >= QT_VERSION_CHECK(5, 4, 0))
-                QStandardPaths::AppLocalDataLocation
-#else
-                QStandardPaths::DataLocation
-#endif
-                ) % '/' % fileName;
+    return QStandardPaths::writableLocation(QStandardPaths::AppLocalDataLocation)
+            % '/' % fileName;
 }
 
-QByteArray MainWindow::loadAccessToken(const QMatrixClient::AccountSettings& account)
+QByteArray MainWindow::loadAccessToken(const AccountSettings& account)
 {
     QFile accountTokenFile { accessTokenFileName(account) };
     if (accountTokenFile.open(QFile::ReadOnly))
@@ -164,7 +305,7 @@ QByteArray MainWindow::loadAccessToken(const QMatrixClient::AccountSettings& acc
     return {};
 }
 
-bool MainWindow::saveAccessToken(const QMatrixClient::AccountSettings& account,
+bool MainWindow::saveAccessToken(const AccountSettings& account,
                                  const QByteArray& accessToken)
 {
     // (Re-)Make a dedicated file for access_token.
@@ -217,6 +358,8 @@ void MainWindow::addConnection(Connection* c, const QString& deviceName)
 {
     Q_ASSERT_X(c, __FUNCTION__, "Attempt to add a null connection");
 
+    using Room = QMatrixClient::Room;
+
     connections.push_back(c);
 
     roomListDock->addConnection(c);
@@ -239,26 +382,102 @@ void MainWindow::addConnection(Connection* c, const QString& deviceName)
         dropConnection(c);
     });
     connect( c, &Connection::networkError, this, [=]{ networkError(c); } );
+    connect( c, &Connection::syncError, this,
+        [this,c] (const QString& message, const QByteArray& details) {
+            QMessageBox msgBox(QMessageBox::Warning, tr("Sync failed"),
+                connections.size() > 1
+                    ? tr("The last sync of account %1 has failed with error: %2")
+                        .arg(c->userId(), message)
+                    : tr("The last sync has failed with error: %1").arg(message),
+                QMessageBox::Retry|QMessageBox::Cancel, this);
+            msgBox.setTextFormat(Qt::PlainText);
+            msgBox.setDefaultButton(QMessageBox::Retry);
+            msgBox.setInformativeText(tr(
+                "Clicking 'Retry' will attempt to resume synchronisation;\n"
+                "Clicking 'Cancel' will stop further synchronisation of this "
+                "account until logout or Quaternion restart."));
+            msgBox.setDetailedText(details);
+            if (msgBox.exec() == QMessageBox::Retry)
+                getNewEvents(c);
+        });
+    using namespace QMatrixClient;
+    connect( c, &Connection::requestFailed, this,
+        [this] (BaseJob* job)
+        {
+            if (job->isBackground())
+                return;
+            auto message = job->error() == BaseJob::UserConsentRequiredError
+                ? tr("Before this server can process your information, you have"
+                     " to agree with its terms and conditions; please click the"
+                     " button below to open the web page where you can do that")
+                : prettyPrint(job->errorString());
+
+            QMessageBox msgBox(QMessageBox::Warning, job->statusCaption(),
+                message, QMessageBox::Close, this);
+            msgBox.setTextFormat(Qt::RichText);
+            msgBox.setDetailedText(
+                "Request URL: " + job->requestUrl().toDisplayString() +
+                "\nResponse:\n" + job->rawData());
+            QPushButton* openUrlButton = nullptr;
+            if (job->errorUrl().isEmpty())
+                msgBox.setDefaultButton(QMessageBox::Close);
+            else
+            {
+                openUrlButton =
+                    msgBox.addButton(tr("Open web page"), QMessageBox::ActionRole);
+                openUrlButton->setDefault(true);
+            }
+            msgBox.exec();
+            if (msgBox.clickedButton() == openUrlButton)
+                QDesktopServices::openUrl(job->errorUrl());
+        });
     connect( c, &Connection::loginError,
              this, [=](const QString& msg){ loginError(c, msg); } );
-    connect( c, &Connection::newRoom, systemTray, &SystemTray::newRoom );
-    connect( c, &Connection::aboutToDeleteRoom,
-             this, [this] (QMatrixClient::Room* r)
-    {
-        if (currentRoom == r)
-            selectRoom(nullptr);
-    });
+    connect( c, &Connection::newRoom, systemTrayIcon, &SystemTrayIcon::newRoom );
+    connect( c, &Connection::createdRoom, this, &MainWindow::selectRoom);
+    connect( c, &Connection::joinedRoom, this, [this] (Room* r, Room* prev)
+        {
+            if (currentRoom == prev)
+                selectRoom(r);
+        });
+    connect( c, &Connection::directChatAvailable, this, &MainWindow::selectRoom);
+    connect( c, &Connection::aboutToDeleteRoom, this,
+             [this] (Room* r) {
+                 if (currentRoom == r)
+                    selectRoom(nullptr);
+             });
 
-    const QString logoutCaption = deviceName.isEmpty() ?
-                tr("Logout %1").arg(c->userId()) :
-                tr("Logout %1/%2").arg(c->userId(), deviceName);
-    const auto logoutAction = new QAction(logoutCaption, c);
-    connectionMenu->insertAction(accountListGrowthPoint, logoutAction);
-    connect( logoutAction, &QAction::triggered, this, [=]{ logout(c); } );
-    connect( c, &Connection::destroyed, this, [=]
+    QString accountCaption = c->userId();
+    if (!deviceName.isEmpty())
+        accountCaption += '/' % deviceName;
+    QString menuCaption = accountCaption;
+    if (connections.size() < 10)
+        menuCaption.prepend('&' % QString::number(connections.size()) % ' ');
+    auto accountMenu = new QMenu(menuCaption, connectionMenu);
+    accountMenu->addAction(tr("Show &access token"), this, [=]
     {
-        connectionMenu->removeAction(logoutAction);
-    } );
+        const QString aToken = c->accessToken();
+        auto accountTokenBox = new QMessageBox(QMessageBox::Information,
+            tr("Access token for %1").arg(accountCaption),
+            tr("Your access token is %1...%2;"
+               " click \"Show details...\" for the full token")
+               .arg(aToken.left(5), aToken.right(5)),
+            QMessageBox::Close, this, Qt::Dialog);
+        accountTokenBox->setModal(false);
+        accountTokenBox->setTextInteractionFlags(
+                    Qt::TextSelectableByKeyboard|Qt::TextSelectableByMouse);
+        accountTokenBox->setDetailedText(aToken);
+        accountTokenBox->setAttribute(Qt::WA_DeleteOnClose);
+        accountTokenBox->show();
+    });
+    accountMenu->addAction(tr("&Logout"), this, [=] { logout(c); });
+    auto menuAction =
+            connectionMenu->insertMenu(accountListGrowthPoint, accountMenu);
+    connect( c, &Connection::destroyed, connectionMenu, [this, menuAction]
+    {
+        connectionMenu->removeAction(menuAction);
+    });
+    createRoomAction->setEnabled(true);
 
     getNewEvents(c);
 }
@@ -270,13 +489,8 @@ void MainWindow::dropConnection(Connection* c)
     if (currentRoom && currentRoom->connection() == c)
         selectRoom(nullptr);
     connections.removeOne(c);
-#if (QT_VERSION >= QT_VERSION_CHECK(5, 4, 0))
     logoutOnExit.removeOne(c);
-#else
-    const auto i = logoutOnExit.indexOf(c);
-    if (i >= 0)
-        logoutOnExit.remove(i);
-#endif
+    createRoomAction->setDisabled(connections.isEmpty());
 
     Q_ASSERT(!connections.contains(c) && !logoutOnExit.contains(c) &&
              !c->syncJob());
@@ -297,14 +511,14 @@ void MainWindow::showLoginWindow(const QString& statusMessage)
     if( dialog.exec() )
     {
         auto connection = dialog.releaseConnection();
-        QMatrixClient::AccountSettings account(connection->userId());
+        AccountSettings account(connection->userId());
         account.setKeepLoggedIn(dialog.keepLoggedIn());
+        account.clearAccessToken(); // Drop the legacy - just in case
         if (dialog.keepLoggedIn())
         {
             account.setHomeserver(connection->homeserver());
             account.setDeviceId(connection->deviceId());
             account.setDeviceName(dialog.deviceName());
-            account.clearAccessToken(); // Drop the legacy - just in case
             if (!saveAccessToken(account, connection->accessToken()))
                 qWarning() << "Couldn't save access token";
         } else
@@ -319,9 +533,9 @@ void MainWindow::showLoginWindow(const QString& statusMessage)
 void MainWindow::invokeLogin()
 {
     using namespace QMatrixClient;
-    SettingsGroup settings("Accounts");
+    const auto accounts = SettingsGroup("Accounts").childGroups();
     bool autoLoggedIn = false;
-    for(const auto& accountId: settings.childGroups())
+    for(const auto& accountId: accounts)
     {
         AccountSettings account { accountId };
         if (!account.homeserver().isEmpty())
@@ -354,13 +568,12 @@ void MainWindow::invokeLogin()
     if (autoLoggedIn)
         showFirstSyncIndicator();
     else
-        QTimer::singleShot(0, this, SLOT(showLoginWindow()));
+        showLoginWindow(tr("Welcome to Quaternion"));
 }
 
 void MainWindow::loginError(Connection* c, const QString& message)
 {
     Q_ASSERT_X(c, __FUNCTION__, "Login error on a null connection");
-    // FIXME: Make ConnectionManager instead of such hacks
     emit c->loggedOut(); // Short circuit login error to logged-out event
     showLoginWindow(message);
 }
@@ -369,32 +582,40 @@ void MainWindow::logout(Connection* c)
 {
     Q_ASSERT_X(c, __FUNCTION__, "Logout on a null connection");
 
-    QMatrixClient::AccountSettings account { c->userId() };
-    QFile(accessTokenFileName(account)).remove();
+    QFile(accessTokenFileName(AccountSettings(c->userId()))).remove();
 
     c->logout();
 }
 
-void MainWindow::selectRoom(QuaternionRoom* r)
+void MainWindow::selectRoom(QMatrixClient::Room* r)
 {
-    currentRoom = r;
+    QElapsedTimer et; et.start();
+    currentRoom = static_cast<QuaternionRoom*>(r);
     setWindowTitle(r ? r->displayName() : QString());
-    chatRoomWidget->setRoom(r);
-    userListDock->setRoom(r);
+    chatRoomWidget->setRoom(currentRoom);
+    userListDock->setRoom(currentRoom);
+    roomSettingsAction->setEnabled(r != nullptr);
+    if (r && !isActiveWindow())
+    {
+        show();
+        activateWindow();
+    }
+    qDebug().noquote() << et << "to "
+        << (r ? "select room " + r->canonicalAlias() : "close the room");
 }
 
 QMatrixClient::Connection* MainWindow::chooseConnection()
 {
     Connection* connection = nullptr;
     QStringList names; names.reserve(connections.size());
-    for (auto c: connections)
+    for (auto c: qAsConst(connections))
         names.push_back(c->userId());
     const auto choice = QInputDialog::getItem(this,
             tr("Choose the account to join the room"), "", names, -1, false);
     if (choice.isEmpty())
         return nullptr;
 
-    for (auto c: connections)
+    for (auto c: qAsConst(connections))
         if (c->userId() == choice)
         {
             connection = c;
@@ -466,39 +687,17 @@ void MainWindow::joinRoom(const QString& roomAlias, Connection* connection)
                                  QMessageBox::Close, QMessageBox::Close);
     }
 
-    using QMatrixClient::JoinRoomJob;
-    auto job = connection->joinRoom(room);
+    using QMatrixClient::BaseJob;
+    auto* job = connection->joinRoom(room);
     // Connection::joinRoom() already connected to success() the code that
     // initialises the room in the library, which in turn causes RoomListModel
     // to update the room list. So the below connection to success() will be
     // triggered after all the initialisation have happened.
-    connect(job, &JoinRoomJob::success, this, [=]
+    connect(job, &BaseJob::success, this, [=]
     {
         statusBar()->showMessage(tr("Joined %1 as %2")
                                  .arg(roomAlias, connection->userId()));
     });
-    connect(job, &JoinRoomJob::failure, this, [=] {
-        QMessageBox messageBox(QMessageBox::Warning,
-                               tr("Failed to join room"),
-                               tr("Joining request returned an error"),
-                               QMessageBox::Close, this);
-        messageBox.setWindowModality(Qt::WindowModal);
-        messageBox.setTextFormat(Qt::PlainText);
-        messageBox.setDetailedText(job->errorString());
-        switch (job->error()) {
-            case JoinRoomJob::NotFoundError:
-                messageBox.setText(
-                    tr("Room %1 not found on the server").arg(roomAlias));
-                break;
-            case JoinRoomJob::IncorrectRequestError:
-                messageBox.setText(
-                    tr("Incorrect id or alias: %1").arg(roomAlias));
-                break;
-            default:
-                messageBox.setText(tr("Error joining %1: %2"));
-        }
-        messageBox.exec();
-    });
 }
 
 void MainWindow::getNewEvents(Connection* c)
@@ -536,7 +735,7 @@ void MainWindow::networkError(Connection* c)
     auto timer = new QTimer(this);
     timer->start(1000);
     showMillisToRecon(c);
-    timer->connect(timer, &QTimer::timeout, [=] {
+    timer->connect(timer, &QTimer::timeout, this, [=] {
         if (c->millisToReconnect() > 0)
             showMillisToRecon(c);
         else
@@ -547,15 +746,66 @@ void MainWindow::networkError(Connection* c)
     });
 }
 
+void MainWindow::sslErrors(QNetworkReply* reply, const QList<QSslError>& errors)
+{
+    for (const auto& error: errors)
+    {
+        if (error.error() == QSslError::NoSslSupport)
+        {
+            static bool showMsgBox = true;
+            if (showMsgBox)
+            {
+                QMessageBox msgBox(QMessageBox::Critical, tr("No SSL support"),
+                                   error.errorString(), QMessageBox::Close, this);
+                msgBox.setInformativeText(
+                    tr("Your SSL configuration does not allow Quaternion"
+                       " to establish secure connections."));
+                msgBox.exec();
+                showMsgBox = false;
+            }
+            return;
+        }
+        QMessageBox msgBox(QMessageBox::Warning, tr("SSL error"),
+                           error.errorString(),
+                           QMessageBox::Abort|QMessageBox::Ignore, this);
+        if (!error.certificate().isNull())
+            msgBox.setDetailedText(error.certificate().toText());
+        if (msgBox.exec() == QMessageBox::Abort)
+            return;
+        NetworkAccessManager::instance()->addIgnoredSslError(error);
+    }
+    reply->ignoreSslErrors(errors);
+}
+
+void MainWindow::proxyAuthenticationRequired(const QNetworkProxy&,
+                                             QAuthenticator* auth)
+{
+    Dialog authDialog(tr("Proxy needs authentication"), this,
+                      Dialog::NoStatusLine, tr("Authenticate"),
+                      Dialog::NoExtraButtons);
+    auto layout = authDialog.addLayout<QFormLayout>();
+    auto userEdit = new QLineEdit;
+    layout->addRow(tr("User name"), userEdit);
+    auto pwdEdit = new QLineEdit;
+    pwdEdit->setEchoMode(QLineEdit::Password);
+    layout->addRow(tr("Password"), pwdEdit);
+
+    if (authDialog.exec() == QDialog::Accepted)
+    {
+        auth->setUser(userEdit->text());
+        auth->setPassword(pwdEdit->text());
+    }
+}
+
 void MainWindow::closeEvent(QCloseEvent* event)
 {
-    for (auto c: connections)
+    for (auto c: qAsConst(connections))
     {
         c->saveState();
         c->stopSync(); // Instead of deleting the connection, merely stop it
 //        dropConnection(c);
     }
-    for (auto c: logoutOnExit)
+    for (auto c: qAsConst(logoutOnExit))
         c->logout(); // For the record, dropConnection() does it automatically
     saveSettings();
     event->accept();
